﻿//---------------------------------------------------------------------------- 
//
//  Copyright (C) CSharp Labs.  All rights reserved.
// 
//  Permission is hereby granted, free of charge, to any person obtaining a copy
//  of this software and associated documentation files (the "Software"), to deal
//  in the Software without restriction, including without limitation the rights
//  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
//  copies of the Software, and to permit persons to whom the Software is
//  furnished to do so, subject to the following conditions:
// 
//  The above copyright notice and this permission notice shall be included in
//  all copies or substantial portions of the Software.
// 
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
//  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
//  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
//  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
//  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
//  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
//  THE SOFTWARE.
// 
// History
//  05/22/13    Created 
//
//---------------------------------------------------------------------------

namespace System.Drawing
{
    using System.Drawing.Imaging;
    using System.Security.Permissions;

    /// <summary>
    /// The BitmapPixels is used for fast access to a Bitmap's Pixels.
    /// </summary>
    public sealed class BitmapPixels : IDisposable
    {
        #region Fields
        /// <summary>
        /// Defines the base Bitmap of the BitmapPixels.
        /// </summary>
        private Bitmap _Bitmap;
        /// <summary>
        /// Defines the BitmapData used to lock and unlock the Bitmap.
        /// </summary>
        private BitmapData _BitmapData;
        /// <summary>
        /// A pointer to the color data of the base Bitmap.
        /// </summary>
        private unsafe RGBAColor* _Pointer = null;
        /// <summary>
        /// Determines if the Bitmap is locked.
        /// </summary>
        private bool _isLocked;
        /// <summary>
        /// Defines the area to lock the Bitmap.
        /// </summary>
        private Rectangle _Bounds;
        /// <summary>
        /// Determines if the bitmap is owned by this instance and should be disposed.
        /// </summary>
        private bool _OwnsBitmap;
        #endregion

        #region Properties
        /// <summary>
        /// Returns the Bitmap's underlying pointer.
        /// </summary>
        public unsafe RGBAColor* Pointer
        {
            get
            {
                return _Pointer;
            }
        }

        /// <summary>
        /// Gets the width of the Bitmap.
        /// </summary>
        public int Width
        {
            get
            {
                return _Bounds.Width;
            }
        }

        /// <summary>
        /// Gets the height of the Bitmap.
        /// </summary>
        public int Height
        {
            get
            {
                return _Bounds.Height;
            }
        }

        /// <summary>
        /// Get the size of the Bitmap.
        /// </summary>
        public Size Size
        {
            get
            {
                return _Bounds.Size;
            }
        }

        /// <summary>
        /// Get the Bitmap this represents.
        /// </summary>
        public Bitmap Bitmap
        {
            get
            {
                return _Bitmap;
            }
        }

        /// <summary>
        /// Determines if the Bitmap is locked in memory. Call Unlock to unlock the Bitmap.
        /// </summary>
        public bool IsLocked
        {
            get
            {
                return _isLocked;
            }
        }

        /// <summary>
        /// Gets or sets the pixel at the specified coordinates.
        /// </summary>
        /// <param name="x">The x axis index.</param>
        /// <param name="y">The y axis index.</param>
        /// <returns>The color at the specified coordinates.</returns>
        public RGBAColor this[int x, int y]
        {
            get
            {
                ThrowIfNotLocked();
                CheckPoint(x, y);

                unsafe
                {
                    return *(_Pointer + _Bounds.Width * y + x);
                }
            }
            set
            {
                ThrowIfNotLocked();
                CheckPoint(x, y);

                unsafe
                {
                    *(_Pointer + _Bounds.Width * y + x) = value;
                }
            }
        }

        /// <summary>
        /// Gets or sets the pixel at the specified coordinates.
        /// </summary>
        /// <param name="p">The pixel coordinates.</param>
        /// <returns>The color at the specified coordinates.</returns>
        public RGBAColor this[Point p]
        {
            get
            {
                return this[p.X, p.Y];
            }
            set
            {
                this[p.X, p.Y] = value;
            }
        }

        /// <summary>
        /// Gets or sets an array of colors.
        /// </summary>
        public RGBAColor[,] Colors
        {
            get
            {
                ThrowIfNotLocked();

                RGBAColor[,] data = new RGBAColor[Size.Width, Size.Height];

                unsafe
                {
                    RGBAColor* colors_ps = _Pointer;

                    for (int y = 0; y < Size.Height; y++)
                    {
                        for (int x = 0; x < Size.Width; x++)
                        {
                            data[x, y] = *colors_ps;
                            colors_ps++;
                        }
                    }
                }

                return data;
            }
            set
            {
                ThrowIfNotLocked();

                if (value.GetLength(0) != Size.Width)
                    throw new ArgumentException("Dimension 0 has an invalid number of elements.");
                if (value.GetLength(1) != Size.Height)
                    throw new ArgumentException("Dimension 1 has an invalid number of elements.");

                unsafe
                {
                    RGBAColor* colors_ps = _Pointer;

                    for (int y = 0; y < Size.Height; y++)
                    {
                        for (int x = 0; x < Size.Width; x++)
                        {
                            *colors_ps = value[x, y];
                            colors_ps++;
                        }
                    }
                }
            }
        }

        #region Property Validation Methods
        private void ThrowIfNotLocked()
        {
            if (!_isLocked)
                throw new InvalidOperationException("Call BitmapPixels.Lock before attempting this operation.");
        }

        /// <summary>
        /// Validates the point.
        /// </summary>
        /// <param name="x">The x axis index.</param>
        /// <param name="y">The y axis index.</param>
        private void CheckPoint(int x, int y)
        {
            if (x < 0)
                throw new ArgumentException("x cannot be less than 0.", "x");
            if (y < 0)
                throw new ArgumentException("y cannot be less than 0.", "y");
            if (x >= _Bounds.Width)
                throw new ArgumentException("x cannot be greater than the Width - 1.", "x");
            if (y >= _Bounds.Height)
                throw new ArgumentException("x cannot be greater than the Height - 1.", "y");
        } 
        #endregion
        #endregion

        #region Constuctor
        /// <summary>
        /// Initializes the BitmapPixels class with a source Bitmap.
        /// </summary>
        /// <param name="source">The source Bitmap to load the pixels from.</param>
        public BitmapPixels(Bitmap source)
        {
            if (source == null)
                throw new ArgumentNullException("source");

            PixelFormat format = source.PixelFormat;

            if (!ValidPixelFormat(format))
                throw new InvalidPixelFormatException(format);

            GraphicsUnit unit = GraphicsUnit.Pixel;

            _OwnsBitmap = false;
            _Bitmap = source;
            _Bounds = Rectangle.Truncate(_Bitmap.GetBounds(ref unit));
        }

        /// <summary>
        /// Initializes the BitmapPixels using an empty Bitmap.
        /// </summary>
        /// <param name="size">The size of the Bitmap.</param>
        public BitmapPixels(Size size)
        {
            _Bitmap = new Bitmap(size.Width, size.Height);
            _Bounds = new Rectangle(Point.Empty, size);
            _OwnsBitmap = true;
        }
        #endregion

        #region Lock / Unlock
        /// <summary>
        /// Lock the Bitmap to access pixel's from pointer. You must unlock the Bitmap or dispose to release resources.
        /// </summary>
        [SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.UnmanagedCode)]
        public void Lock()
        {
            if (_isLocked)
                return;

            _BitmapData = _Bitmap.LockBits(_Bounds, ImageLockMode.ReadWrite, PixelFormat.Format32bppArgb);
            _isLocked = true;

            unsafe
            {
                _Pointer = (RGBAColor*)_BitmapData.Scan0.ToPointer();
            }
        }

        /// <summary>
        /// Unlocks the Bitmap after a call to Lock.
        /// </summary>
        [SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.UnmanagedCode)]
        public void Unlock()
        {
            if (!_isLocked)
                return;

            _Bitmap.UnlockBits(_BitmapData);
            _isLocked = false;

            unsafe
            {
                _Pointer = null;
            }
        }
        #endregion

        #region Disposing
        /// <summary>
        /// Destructor to handle disposing this instance.
        /// </summary>
        ~BitmapPixels()
        {
            Dispose(false);
        }
        /// <summary>
        /// Disposes the instance, releasing managed and unmanaged memory.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Disposes unmanaged and optionally, managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        private void Dispose(bool disposing)
        {
            if (_isLocked)
                Unlock();

            if (disposing)
            {
                if (_Bitmap != null && _OwnsBitmap)
                {
                    _Bitmap.Dispose();
                    _Bitmap = null;
                }
            }
        }
        #endregion

        /// <summary>
        /// Determines if the specified PixelFormat is a format which can be loaded with the BitmapPixels class.
        /// </summary>
        /// <param name="format">The format of the color data for each pixel in the image.</param>
        /// <returns>true if the format is 24bpp Rgb or 32bpp Argb/PArgb/Rgb; otherwise false.</returns>
        public static bool ValidPixelFormat(PixelFormat format)
        {
            switch (format)
            {
                case PixelFormat.Format24bppRgb:
                case PixelFormat.Format32bppArgb:
                case PixelFormat.Format32bppPArgb:
                case PixelFormat.Format32bppRgb:
                    return true;
                default:
                    return false;
            }
        }
    }
}
